import createError from 'http-errors'
import {
  GraphQLSchema,
  validateSchema,
  Source,
  parse,
  validate,
  execute as executeGraphQL,
  formatError
} from 'graphql'
export const errorHandler = () => async (ctx, next) => {
  try {
    await next()
  } catch (error) {
    let httpError = createError(error)
    if (!httpError.expose) httpError = createError()
    ctx.response.status = httpError.status
    if (!ctx.response.body) ctx.response.body = {}
    ctx.response.body.errors = Array.isArray(httpError.graphqlErrors)
      ? httpError.graphqlErrors.map(formatError)
      : [formatError(httpError)]
    ctx.app.emit('error', error, ctx)
  }
}
export const execute = options => {
  if (typeof options === 'undefined')
    throw createError('GraphQL execute middleware options missing.')
  if (!isPlainObject(options))
    throw createError('GraphQL execute middleware options must be an object.')
  const ALLOWED_OPTIONS = [
    'schema',
    'rootValue',
    'contextValue',
    'fieldResolver',
    'override'
  ]
  checkOptions(options, ALLOWED_OPTIONS, 'GraphQL execute middleware')
  if (typeof options.schema !== 'undefined') checkSchema(options.schema)
  if (
    typeof options.override !== 'undefined' &&
    typeof options.override !== 'function'
  )
    throw createError(
      'GraphQL execute middleware `override` option must be a function.'
    )
  return async (ctx, next) => {
    if (typeof ctx.request.body === 'undefined')
      throw createError('Request body missing.')
    if (!isPlainObject(ctx.request.body))
      throw createError(400, 'Request body must be a JSON object.')
    if (!('query' in ctx.request.body))
      throw createError(400, 'GraphQL operation field `query` missing.')
    let document

    try {
      document = parse(new Source(ctx.request.body.query))
    } catch (error) {
      throw createError(400, `GraphQL query syntax error: ${error.message}`)
    }

    let optionsOverride = {}

    if (options.override) {
      optionsOverride = await options.override(ctx)
      if (!isPlainObject(optionsOverride))
        throw createError(
          'GraphQL execute middleware options must be an object, or an object promise.'
        )
      checkOptions(
        optionsOverride,
        ALLOWED_OPTIONS.filter(option => option !== 'override'),
        'GraphQL execute middleware `override` option return'
      )
      if (typeof optionsOverride.schema !== 'undefined')
        checkSchema(optionsOverride.schema)
    }

    const execute = {
      ...options,
      ...optionsOverride
    }
    const queryValidationErrors = validate(execute.schema, document)
    if (queryValidationErrors.length)
      throw createError(400, 'GraphQL query validation errors.', {
        graphqlErrors: queryValidationErrors
      })
    let result

    try {
      result = await executeGraphQL({
        ...execute,
        document,
        variableValues: ctx.request.body.variables,
        operationName: ctx.request.body.operationName
      })
    } catch (error) {
      throw createError(
        400,
        `GraphQL operation field invalid: ${error.message}`
      )
    }

    if (result.data)
      ctx.body = {
        data: result.data
      }
    if (result.errors)
      throw createError(200, 'GraphQL errors.', {
        graphqlErrors: result.errors
      })
    ctx.response.status = 200
    await next()
  }
}

const isPlainObject = value =>
  typeof value === 'object' && value !== null && !Array.isArray(value)

const checkOptions = (options, allowed, description) => {
  const invalid = Object.keys(options).filter(
    option => !allowed.includes(option)
  )
  if (invalid.length)
    throw createError(
      `${description} options invalid: \`${invalid.join('`, `')}\`.`
    )
}

const checkSchema = schema => {
  if (!(schema instanceof GraphQLSchema))
    throw createError(
      'GraphQL schema is required and must be a `GraphQLSchema` instance.'
    )
  const schemaValidationErrors = validateSchema(schema)
  if (schemaValidationErrors.length)
    throw createError('GraphQL schema validation errors.', {
      graphqlErrors: schemaValidationErrors
    })
}
